import os.path
import struct

from .memory import Memory


def read_file(file):
    with open(file, "rb") as fp:
        return fp.read()


def load_mzt(file):
    data = read_file(file)
    if data[0] != 1:
        print(f"invalid MZT file: {file}, : Attribute is ${data[0]:02X}")
        exit()
    size, offset, start = struct.unpack("<3H", data[0x12:0x18])
    if size + 128 != len(data):
        print(f"invalid MZT file: {file}, may be multiple data block")
        exit()

    return Memory(data[128 : size + 128], offset=offset, start=start)


def load_bin(file, offset):
    data = read_file(file)
    return Memory(data, offset=offset)


def load(file, offset):
    _, ext = os.path.splitext(file)
    ext = ext.lower()
    if ext == ".mzt":
        return load_mzt(file)
    else:
        return load_bin(file, offset)



import filecmp
import io
import subprocess
from contextlib import redirect_stdout
from pathlib import Path

from src.yad80.cli import cli_main

ASM = r"C:\Programs\z88dk\bin\z80asm.exe"
BASE = Path(__file__).parent
TEMP_SRC = "temp.asm"
TEMP_OUT = "temp.bin"


def unlink_temp():
    for path in Path(".").glob("temp.*"):
        path.unlink(True)


def test_rom():
    for file in BASE.glob("*.rom"):
        unlink_temp()
        target = BASE / file

        fp = io.StringIO()
        with redirect_stdout(fp):
            cli_main(f"-e -c 0-79 -- {target}".split())
            Path(TEMP_SRC).write_text(fp.getvalue())

        assert True

        result = subprocess.run([ASM, "-b", TEMP_SRC], stdout=True)
        assert result.returncode == 0, f"asm {TEMP_SRC} of {file}"

        assert filecmp.cmp(target, TEMP_OUT), f"compare {target} with {TEMP_OUT}"
    unlink_temp()


def test_MZT():
    for file in BASE.glob("*.mzt"):
        unlink_temp()
        target = BASE / file

        fp = io.StringIO()
        with redirect_stdout(fp):
            cli_main(f"-e {target}".split())
            Path(TEMP_SRC).write_text(fp.getvalue())

        result = subprocess.run([ASM, "-b", TEMP_SRC], stdout=True)
        assert result.returncode == 0, f"asm {TEMP_SRC} of {file}"

        assert (
            target.read_bytes()[128:] == Path(TEMP_OUT).read_bytes()
        ), f"compare {target} with {TEMP_OUT}"
    unlink_temp()


def test_asm():
    for file in BASE.glob("*.bin"):
        unlink_temp()

        target = BASE / file
        fp = io.StringIO()
        with redirect_stdout(fp):
            cli_main(f"-m 65536 {target}".split())
            Path(TEMP_SRC).write_text(fp.getvalue())

        result = subprocess.run([ASM, "-b", TEMP_SRC], stdout=True)
        assert result.returncode == 0, f"asm {TEMP_SRC} of {file}"

        assert (
            target.read_bytes() == Path(TEMP_OUT).read_bytes()
        ), f"compare {target} with {TEMP_OUT}"
    unlink_temp()



from .exceptions import InstructionError
from .mnemonic_cb import MNEMONIC_CB
from .mnemonic_dd_fd import MNEMONIC_DD_FD
from .mnemonic_defs import (
    ARITHMETIC,
    CC,
    REG8,
    REG16_AF,
    REG16_SP,
    ROTATE_SHIFT,
    uint8_to_int8,
)
from .mnemonic_ed import MNEMONIC_ED


def ld_reg8_reg8(op, _):
    r1 = (op >> 3) & 7
    r2 = op & 7
    if r1 == r2 == 6:  # just in case
        return "HALT"
    return f"LD {REG8[r1]},{REG8[r2]}"


def ld_reg8_n(op, mem):
    r = (op >> 3) & 7
    n = mem.next_byte()
    return f"LD {REG8[r]},${n:02X}"


def ld_reg16_nn(op, mem):
    rr = (op >> 4) & 3
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD {REG16_SP[rr]},${n2:02X}{n1:02X}"


def ld_mem_HL(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD (${n2:02X}{n1:02X}),HL"


def ld_HL_mem(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD HL,(${n2:02X}{n1:02X})"


def ld_mem_A(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD (${n2:02X}{n1:02X}),A"


def ld_A_mem(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD A,(${n2:02X}{n1:02X})"


def arithmetic_reg8(op, _):
    p = (op >> 3) & 7
    r = op & 7
    return f"{ARITHMETIC[p]}{REG8[r]}"


def arithmetic_reg8_n(op, mem):
    p = (op >> 3) & 7
    n = mem.next_byte()
    return f"{ARITHMETIC[p]}${n:02X}"


def add_hl(op, _):
    rr = (op >> 4) & 3
    return f"ADD HL,{REG16_SP[rr]}"


def inc_reg16(op, _):
    rr = (op >> 4) & 3
    return f"INC {REG16_SP[rr]}"


def dec_reg16(op, _):
    rr = (op >> 4) & 3
    return f"DEC {REG16_SP[rr]}"


def inc_reg8(op, _):
    r = (op >> 3) & 7
    return f"INC {REG8[r]}"


def dec_reg8(op, _):
    r = (op >> 3) & 7
    return f"DEC {REG8[r]}"


def rotate_shift(op, _):
    p = (op >> 3) & 7
    return ROTATE_SHIFT[p]


def djnz(_, mem):
    n = mem.next_byte()
    addr = mem.addr + uint8_to_int8(n)
    return f"DJNZ ${addr:04X}"


def jr(_, mem):
    n = mem.next_byte()
    addr = mem.addr + uint8_to_int8(n)
    return f"JR ${addr:04X}"


def jr_cc(op, mem):
    cc = (op >> 3) & 7 - 4
    n = mem.next_byte()
    addr = mem.addr + uint8_to_int8(n)
    return f"JR {CC[cc]},${addr:04X}"


def ret_cc(op, _):
    cc = (op >> 3) & 7
    return f"RET {CC[cc]}"


def pop_reg16(op, _):
    rr = (op >> 4) & 3
    return f"POP {REG16_AF[rr]}"


def push_reg16(op, _):
    rr = (op >> 4) & 3
    return f"PUSH {REG16_AF[rr]}"


def jp_cc(op, mem):
    cc = (op >> 3) & 7
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"JP {CC[cc]},${n2:02X}{n1:02X}"


def jp(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"JP ${n2:02X}{n1:02X}"


def call_cc(op, mem):
    cc = (op >> 3) & 7
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"CALL {CC[cc]},${n2:02X}{n1:02X}"


def call(_, mem):
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"CALL ${n2:02X}{n1:02X}"


def rst(op, _):
    p = (op >> 3) & 7
    return f"RST ${p * 8:02X}"


def out(_, mem):
    n = mem.next_byte()
    return f"OUT (${n:02X}),A"


def in_(_, mem):
    n = mem.next_byte()
    return f"IN A,(${n:02X})"


def opecode_cb(_, mem):
    op = mem.next_byte()
    return MNEMONIC_CB[op](op, mem)


def opecode_ed(_, mem):
    op = mem.next_byte()
    func = MNEMONIC_ED.get(op)
    if func is None:
        raise InstructionError(f"invalid instruction ed {op:02x}")
    return func(op, mem)


def opecode_dd_fd(op1, mem):
    op2 = mem.next_byte()
    # print(f"{op2:02x} ", end="")
    func = MNEMONIC_DD_FD.get(op2)
    if func is None:
        raise InstructionError(f"invalid instruction {op1:02x} {op2:02x}")
    return func(op1, op2, mem)


MNEMONIC = {
    0x00: lambda *_: "NOP",
    0x01: ld_reg16_nn,
    0x02: lambda *_: "LD (BC),A",
    0x03: inc_reg16,
    0x04: inc_reg8,
    0x05: dec_reg8,
    0x06: ld_reg8_n,
    0x07: rotate_shift,
    0x08: lambda *_: "EX AF,AF'",
    0x09: add_hl,
    0x0A: lambda *_: "LD A,(BC)",
    0x0B: dec_reg16,
    0x0C: dec_reg8,
    0x10: djnz,
    0x12: lambda *_: "LD (DE),A",
    0x18: jr,
    0x1A: lambda *_: "LD A,(DE)",
    0x20: jr_cc,
    0x22: ld_mem_HL,
    0x2A: ld_HL_mem,
    0x32: ld_mem_A,
    0x3A: ld_A_mem,
    0x40: ld_reg8_reg8,
    0x76: lambda *_: "HALT",
    0x80: arithmetic_reg8,
    0xC0: ret_cc,
    0xC1: pop_reg16,
    0xC2: jp_cc,
    0xC3: jp,
    0xC4: call_cc,
    0xC5: push_reg16,
    0xC6: arithmetic_reg8_n,
    0xC7: rst,
    0xC9: lambda *_: "RET",
    0xCB: opecode_cb,
    0xCD: call,
    0xD3: out,
    0xD9: lambda *_: "EXX",
    0xDB: in_,
    0xE3: lambda *_: "EX (SP),HL",
    0xE9: lambda *_: "JP (HL)",
    0xEB: lambda *_: "EX DE,HL",
    0xF3: lambda *_: "DI",
    0xF9: lambda *_: "LD SP,HL",
    0xFB: lambda *_: "EI",
    0xDD: opecode_dd_fd,
    0xED: opecode_ed,
    0xFD: opecode_dd_fd,
}


def init_instruction_dict():
    for op in range(0x41, 0x80):
        if op == 0x76:  # HALT
            continue
        MNEMONIC[op] = MNEMONIC[0x40]  # LD r,r'

    for op in range(0x81, 0xC0):
        MNEMONIC[op] = MNEMONIC[0x80]  # 8 bit arithmetic

    for n in range(1, len(REG16_SP)):
        MNEMONIC[0x01 + n * 0x10] = MNEMONIC[0x01]  # LD rr,nn
        MNEMONIC[0x03 + n * 0x10] = MNEMONIC[0x03]  # INC rr
        MNEMONIC[0x09 + n * 0x10] = MNEMONIC[0x09]  # ADD HL,rr
        MNEMONIC[0x0B + n * 0x10] = MNEMONIC[0x0B]  # DEC rr

    for n in range(1, len(REG8)):
        MNEMONIC[0x04 + n * 8] = MNEMONIC[0x04]  # INC r
        MNEMONIC[0x05 + n * 8] = MNEMONIC[0x05]  # DEC r
        MNEMONIC[0x06 + n * 8] = MNEMONIC[0x06]  # LD r,n

    for n in range(1, len(ROTATE_SHIFT)):
        MNEMONIC[0x07 + n * 8] = MNEMONIC[0x07]  # rotate and shift

    for n in range(1, 4):
        MNEMONIC[0x20 + n * 8] = MNEMONIC[0x20]  # JR cc

    for n in range(1, len(CC)):
        MNEMONIC[0xC0 + n * 8] = MNEMONIC[0xC0]  # RET cc
        MNEMONIC[0xC2 + n * 8] = MNEMONIC[0xC2]  # JP CC,nnnn
        MNEMONIC[0xC4 + n * 8] = MNEMONIC[0xC4]  # CALL CC

    for n in range(1, len(REG16_AF)):
        MNEMONIC[0xC1 + n * 0x10] = MNEMONIC[0xC1]  # POP rr
        MNEMONIC[0xC5 + n * 0x10] = MNEMONIC[0xC5]  # PUSH rr

    for n in range(1, len(ARITHMETIC)):
        MNEMONIC[0xC6 + n * 8] = MNEMONIC[0xC6]  # 8 bit arithmetic

    for n in range(1, 8):
        MNEMONIC[0xC7 + n * 8] = MNEMONIC[0xC7]  # RST n


init_instruction_dict()



from .disasm import disasm_line, disasm_nlines
from .exceptions import AddressError, InstructionError
from .memory import Memory

__all__ = ["Memory", "disasm_nlines", "disasm_line", "AddressError", "InstructionError"]



from .exceptions import InstructionError
from .mnemonic_defs import (
    ARITHMETIC,
    BIT_OP,
    REG8,
    REG16_SP,
    ROTATE_SHIFT_R,
    uint8_to_int8,
)


def add_reg16(op1, op2, _):
    ixy = "IX" if op1 == 0xDD else "IY"
    rr = (op2 >> 4) & 3
    src = ixy if rr == 2 else REG16_SP[rr]
    return f"ADD {ixy},{src}"


def ld_reg8_indexed(op1, op2, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    ofs = uint8_to_int8(mem.next_byte())
    sign = "-" if ofs < 0 else "+"
    r = (op2 >> 3) & 7
    if r == 6:
        raise InstructionError(f"invalid instruction {op1:02x} {op2:02x}")
    return f"LD {REG8[r]},({ixy}{sign}${abs(ofs):02X})"


def ld_indexed_reg8(op1, op2, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    ofs = uint8_to_int8(mem.next_byte())
    sign = "-" if ofs < 0 else "+"
    r = op2 & 7
    return f"LD ({ixy}{sign}${abs(ofs):02x}),{REG8[r]}"


def ld_index_n(op1, _, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD {ixy},${n2:02X}{n1:02X}"


def ld_mem_index(op1, _, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD (${n2:02X}{n1:02X}),{ixy}"


def ld_index_mem(op1, _, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    n1 = mem.next_byte()
    n2 = mem.next_byte()
    return f"LD {ixy},(${n2:02X}{n1:02X})"


def ld_indexed_n(op1, _, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    ofs = uint8_to_int8(mem.next_byte())
    sign = "-" if ofs < 0 else "+"
    n = mem.next_byte()
    return f"LD ({ixy}{sign}${abs(ofs):02X}),${n:02X}"


def inc_dec_indexed(op1, op2, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    incdec = "INC" if op2 == 0x34 else "DEC"
    ofs = uint8_to_int8(mem.next_byte())
    sign = "-" if ofs < 0 else "+"
    return f"{incdec} ({ixy}{sign}${abs(ofs):02X})"


def bit_shift(op1, _, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    n = mem.next_byte()
    ofs = uint8_to_int8(n)
    sign = "-" if ofs < 0 else "+"
    op2 = mem.next_byte()
    r = op2 & 7
    if r != 6:
        raise InstructionError(f"invalid instruction {op1:02x} cb {n:02x} {op2:02x}")
    mask = (op2 >> 6) & 3
    if mask == 0:
        shift_op = (op2 >> 3) & 7
        return f"{ROTATE_SHIFT_R[shift_op]} ({ixy}{sign}${abs(ofs):02X})"
    else:
        bit_op = (op2 >> 6) & 7
        n = (op2 >> 3) & 7
        return f"{BIT_OP[bit_op]} {n},({ixy}{sign}${abs(ofs):02X})"


def arithmetic_indexed(op1, op2, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    n = mem.next_byte()
    ofs = uint8_to_int8(n)
    sign = "-" if ofs < 0 else "+"
    p = (op2 >> 3) & 7
    return f"{ARITHMETIC[p]} ({ixy}{sign}${abs(ofs):02X})"


def ld_reg8_n(op1, op2, mem):
    ixy = "IX" if op1 == 0xDD else "IY"
    r = (op2 >> 3) & 7
    n = mem.next_byte()
    return f"LD {ixy}{REG8[r]},${n:02X}"


def ld_reg8_reg8(op1, op2, _):
    ixy = "IX" if op1 == 0xDD else "IY"
    r1 = (op2 >> 3) & 7
    r2 = op2 & 7
    if r1 == r2 == 6:  # just in case
        return "NOP"
    reg1 = REG8[r1]
    reg2 = REG8[r2]
    if reg1 in {"H", "L"}:
        reg1 = ixy + reg1
    if reg2 in {"H", "L"}:
        reg2 = ixy + reg2
    return f"LD {reg1},{reg2}"


def arithmetic_reg8(op1, op2, _):
    ixy = "IX" if op1 == 0xDD else "IY"
    p = (op2 >> 3) & 7
    r = op2 & 7
    reg = REG8[r]
    if reg in {"H", "L"}:
        reg = ixy + reg

    return f"{ARITHMETIC[p]}{reg}"


MNEMONIC_DD_FD = {
    0x09: add_reg16,
    0x21: ld_index_n,
    0x22: ld_mem_index,
    0x2A: ld_index_mem,
    0x34: inc_dec_indexed,
    0x35: inc_dec_indexed,
    0x36: ld_indexed_n,
    0x26: ld_reg8_n,  # undocumented
    0x2E: ld_reg8_n,  # undocumented
    0x23: lambda op, *_: "INC IX" if op == 0xDD else "INC IY",
    0x2B: lambda op, *_: "DEC IX" if op == 0xDD else "DEC IY",
    0x24: lambda op, *_: "INC IXH" if op == 0xDD else "INC IYH",  # undocumented
    0x25: lambda op, *_: "DEC IXH" if op == 0xDD else "DEC IYH",  # undocumented
    0x2C: lambda op, *_: "INC IXL" if op == 0xDD else "INC IYL",  # undocumented
    0x2D: lambda op, *_: "DEC IXL" if op == 0xDD else "DEC IYL",  # undocumented
    0x44: ld_reg8_reg8,
    0x46: ld_reg8_indexed,
    0x70: ld_indexed_reg8,
    0x86: arithmetic_indexed,
    0xE1: lambda op, *_: "POP IX" if op == 0xDD else "POP IY",
    0xE5: lambda op, *_: "PUSH IX" if op == 0xDD else "PUSH IY",
    0xE3: lambda op, *_: "EX (SP),IX" if op == 0xDD else "EX (SP),IY",
    0xE9: lambda op, *_: "JP (IX)" if op == 0xDD else "JP (IY)",
    0xF9: lambda op, *_: "LD SP,IX" if op == 0xDD else "LD SP,IY",
    0xCB: bit_shift,
}


def init_instruction_dict():
    for n in range(1, 4):
        MNEMONIC_DD_FD[0x09 + n * 0x10] = MNEMONIC_DD_FD[0x09]  # LD ixy,REG16

    for n in range(1, len(REG8)):
        MNEMONIC_DD_FD[0x46 + n * 8] = MNEMONIC_DD_FD[0x46]  # LD r,(ixy+n)
        MNEMONIC_DD_FD[0x70 + n] = MNEMONIC_DD_FD[0x70]  # LD r,(ixy+n)

    for n in range(1, len(ARITHMETIC)):
        MNEMONIC_DD_FD[0x86 + n * 8] = MNEMONIC_DD_FD[0x86]  # arithmetic

    # undocumented LD r,r'
    for r1 in [0, 1, 2, 3, 4, 5, 7]:
        for r2 in [4, 5]:
            op = 0x40 + (r1 << 3) + r2
            MNEMONIC_DD_FD[op] = ld_reg8_reg8
    for r1 in [4, 5]:
        for r2 in [0, 1, 2, 3, 5, 7]:
            op = 0x40 + (r1 << 3) + r2
            MNEMONIC_DD_FD[op] = ld_reg8_reg8

    # undocumented arithmetic r
    for aop in range(8):
        for r in [4, 5]:
            op = 0x80 + (aop << 3) + r
            MNEMONIC_DD_FD[op] = arithmetic_reg8


init_instruction_dict()



from .exceptions import InstructionError
from .mnemonic_defs import BIT_OP, REG8, ROTATE_SHIFT_R


def rotate_shift_r(op, _):
    p = (op >> 3) & 7
    r = op & 7
    if ROTATE_SHIFT_R[p] is None:
        raise InstructionError(f"invalid instruction cb {op:02x}")
    return f"{ROTATE_SHIFT_R[p]} {REG8[r]}"


def bit_operation(op, _):
    bit_op = op >> 6
    n = (op >> 3) & 7
    r = op & 7
    return f"{BIT_OP[bit_op]} {n},{REG8[r]}"


MNEMONIC_CB = {
    0x00: rotate_shift_r,
    0x40: bit_operation,
}


def init_instruction_hash():
    for n in range(1, 0x40):
        MNEMONIC_CB[n] = MNEMONIC_CB[0]  # rotate r, shift r
    for n in range(0x41, 0x100):
        MNEMONIC_CB[n] = MNEMONIC_CB[0x40]  # bit operation


init_instruction_hash()



from .exceptions import AddressError


class Memory:
    def __init__(self, block, start=0, offset=0):
        self.block = block
        self.offset = offset
        self.current = 0
        self.min_addr = offset
        self.max_addr = offset + len(block) - 1
        if start != 0:
            self.start = start
        else:
            self.start = self.min_addr

    def next_byte(self):
        if self.current < len(self.block):
            b = self.block[self.current]
            self.current += 1
            return b
        return None

    def rewind(self):
        self.current = 0

    def __repr__(self):
        return (
            f"offset:{self.offset:x}, start: {self.start:04x}, "
            f"addr:{self.min_addr:04x}-{self.max_addr:04x}, "
            f"size:{len(self):04x}"
        )

    @property
    def addr(self):
        return self.current + self.offset

    @addr.setter
    def addr(self, value):
        if value < self.min_addr or value > self.max_addr:
            raise AddressError(f"invalid addr {value:04x}")
        self.current = value - self.offset

    def __len__(self):
        return len(self.block)

    def addr_in(self, addr):
        return self.min_addr <= addr <= self.max_addr

    def __getitem__(self, index):
        if isinstance(index, int):
            if index < self.min_addr or index > self.max_addr:
                raise AddressError(f"{index:04x} out of range")
            return self.block[index - self.offset]

        if not isinstance(index, slice):
            raise AddressError(f"invalid index {index!r}")

        start, stop = index.start, index.stop
        if start is None:
            start = self.min_addr

        if stop is None:
            stop = self.max_addr + 1

        if start < self.min_addr or stop > self.max_addr + 1:
            raise AddressError(f"invalid slice ${start}:${stop:04x}")

        return self.block[start - self.offset : stop - self.offset]
